#!/opt/bin/perl5 -w 


#---------------------------------------------------------------------
#
# d2xpp.pl
#
#---------------------------------------------------------------------
#
# preprocessor for dns2xml.pl - 
#
#---------------------------------------------------------------------

use strict; 
use File::Copy;
use Getopt::Std;
use Time::Local;
use Data::Dumper;


#---------------------------------------------------------------------
#
# Initialization
#
#---------------------------------------------------------------------


#
# Process command line arguments
#

my ($inputDir, $outputDir, $debug) = &ProcessCmdLine();

# errorfile will be written to the output directory

my $errorfile       = "Error.log"; 
my $namedConfigFile = "named.conf";


#---------------------------------------------------------------------
#
# Main Processing Loop
#
#---------------------------------------------------------------------

my $zoneList        = &GetZoneList($inputDir, $namedConfigFile);
# print Dumper $zoneList; exit;

# clean out outputdir first? gzip or something?
my @parsedFile;
my %includedfiles;
my $includedLevel = 0;

if (!-e $outputDir) { mkdir($outputDir, 0750) or die "Can't create $outputDir, $!\n";} # what about sub dirs ?
foreach my $zone (@{$zoneList}) {
  @parsedFile = (); 
  %includedfiles=();
  &ParseDBFile("$inputDir/$zone->{FILE}","$zone->{NAME}","");
  open(PPDBFILE,">$outputDir/$zone->{FILE}") or 
    do {print STDERR "Can't output $outputDir/$zone->{FILE}, $!\n";next;};
  print PPDBFILE @parsedFile;
  close PPDBFILE;
} 

copy("$inputDir/$namedConfigFile","$outputDir/$namedConfigFile") or
  do {print STDERR "Can't output $outputDir/$namedConfigFile, $!\n";};

#---------------------------------------------------------------------
#
# ProcessResourceRec - 
#
#---------------------------------------------------------------------

sub ProcessResourceRec() {
  my($owner, $ttl, $class, $type, $rdata, $zoneOrigin, $zoneTTL, $soahost)=@_;
  my $goodWrite = 1; # if $goodWrite becomes 0, don't write out this RR, log as error.
  $ttl = ""; # $ttl not currently utilized, but available for implementation!
# Format for printing out expanded resource records for easy viewing while developing.  
format STDOUT = 
@<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< @<<<<<<<<<< @<<<<  @<<<<<<< @<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$owner, $ttl, $class, $type, $rdata
.
  #
  # Make sure current origin ends in a period.
  #
  if (!($zoneOrigin =~ /\.$/)) { $zoneOrigin = $zoneOrigin . "."; }	     
   
  #
  # "@" expands to value of current origin.
  #
  if ($owner eq "@") { $owner = $zoneOrigin; }

  #
  # Fully expand owners that don't end in a "."
  #
  if (!($owner =~ /\.$/)){ 
      $owner="$owner.$zoneOrigin"; 
  }
  
  #
  # Fully expand rdata that doesn't end in a "."
  #
  if (!($rdata =~ /\.$/)){ 
    if (
	($type eq "CNAME") || 
	($type eq "MX")    || 
	($type eq "NS")
       ) 
      {
	$rdata = "$rdata.$zoneOrigin";
      }
    else {} #don't add anything (A, HINFO, TXT, etc.)
  }
 
  #
  # If SOA, extract and return header info, else print out expanded RR.
  #
  if ($type eq "SOA") {
    push (@parsedFile,"$owner\t$ttl\t$class\t$type\t$rdata\n");
    $debug && print "$owner $ttl $class $type $rdata\n";
    my ($soahost, $mailaddr, $serial, $refresh, $retry, $expire, $minttl) =
      $rdata =~ /^(\S+)\s+(\S+)\s+\D*(\d+)\D*(\d+)\D*(\d+)\D*(\d+)\D*(\d+)\D*\)?/;
    return ($soahost,$mailaddr,$serial,$refresh,$retry,$expire,$minttl);
  } 
  else {
    if ($goodWrite) {
      push (@parsedFile,"$owner\t$ttl\t$class\t$type\t$rdata\n");
      $debug && write;
    }
    return 0;
  }
}


#---------------------------------------------------------------------
#
# ParseDBFile - 
#
#---------------------------------------------------------------------

sub ParseDBFile() {
  my ($dbFile, $zoneOrigin, $zoneTTL) = @_;
  my ($owner, $class, $class_ttl, $ttl, $type, $rdata); 
  my ($lastOwner, $lastClass, @tmp, $currArg, $inside_soa);
  my ($soahost, $soamailaddr, $soaserial, $soarefresh, $soaretry, $soaexpire, $soaminttl);
  $lastClass="";
  $inside_soa=0;

  my $vals=0;
  my $soastring = "";
  print "---------------------------------------- Parsing $dbFile\n";
  open(DBFILE,"$dbFile") or return 0; #skips files can't open. Write to error log?
  my @dbfile = <DBFILE>;
  close DBFILE;
  
  while (@dbfile) {
    $_=shift(@dbfile);
    next if (/^;/);
    s/\s?;.*//;

    #
    # Process a Resource Record
    #

    # This needs to be generalized to recognize ALL rr types!!!
    if (/^(\S*)?\s+(.*)\s+(A|PTR|CNAME|MX|TXT|HINFO|NS|SOA)\s+(.*)/) {
      
      # Resource Record type      
      $type  = $3; 
      
      # Resource Record data      
      $rdata = $4;
      
      # Resource Record owner      
      $owner = $1; 

      if ($owner =~ /\*/) {print "$owner - Wildcards currently not supported by dns2xml\n"; next;}

      if ($owner eq "") {
	$owner = $lastOwner;
      } else {
	$lastOwner = $owner;
      }
      
      # Resource Record class and ttl      
      # Go through some hoops because can have class ttl, ttl class, 
      # one or neither.
      # A bit moot, since we're not going to use ttl anyway...

      if ($2 ne "") {@tmp = split(/\s+/,$2)};

      if (@tmp) {$currArg = pop(@tmp)} else {$currArg="";} 
      
      if ($currArg =~ m/\d/) {
	$ttl = $currArg; 
	if (@tmp) {
	  $class = pop(@tmp); 
	  $lastClass = $class;
	} else {
	  $class = $lastClass;
	}
      } 
      elsif ($currArg ne "") {
	$class = $currArg; 
	$lastClass = $class;
	if (@tmp) {
	  $ttl = pop(@tmp);
	} else {
	  $ttl = $zoneTTL;
	}
      }
      else {
	$class = $lastClass;
	$ttl   = $zoneTTL;
      }

      if ($type ne "SOA")  { 
	&ProcessResourceRec($owner, $ttl, $class, $type, $rdata, $zoneOrigin, $zoneTTL, $soahost);
      }
      else {   
	# Process the SOA RR a little differently, so flag that we're in it  
	$inside_soa=1; 
      }
    }
    
    # (Messy) way to process multi-line SOA records. TODO - explain how works.
    if ($inside_soa) {
      my $soadata = $_;
      chomp($soadata);
      $soadata =~ s/\(//;
      $soadata =~ s/\;.*/ /; # lose comments

      if ($soadata =~ /\)/) {
	$inside_soa = 0;
	$soadata =~ s/\)//;
      }

      $soastring = $soastring . $soadata;

      if (!($inside_soa)) {
	$soastring =~ s/.*SOA//;
	$soastring =~ s/\s+/ /g;
	$soastring =~ s/^\s+//g;
	($soahost, $soamailaddr, $soaserial, $soarefresh, $soaretry, $soaexpire, $soaminttl) =
	  &ProcessResourceRec($owner, $ttl, $class, $type, $soastring, $zoneOrigin, $zoneTTL, $soahost);
	$zoneTTL=$soaminttl;
      }  
    }
    
    #
    # Process a DNS dbfile directive 
    #
    elsif (/^\$(ORIGIN|INCLUDE|TTL)/) {
      chomp;
      my @directive = split /\s+/;
      &ProcessDirective(\@directive, \$zoneOrigin, \$zoneTTL);
    }
  }
}


#---------------------------------------------------------------------
#
# ProcessDirective - Handle $ORIGIN, $INCLUDE, $TTL
#
#---------------------------------------------------------------------

sub ProcessDirective() {
  my ($directive, $zoneOrigin, $zoneTTL) = @_;
  
  my $command = shift(@{$directive});
  
  #
  # $ORIGIN 
  #
  if ($command eq '$ORIGIN') {
    # modifying actual $zoneOrigin from ParseDBFile, not a copy.
    $$zoneOrigin = shift(@{$directive}); 

    $debug && do {
      print "\$ORIGIN      - new origin is $$zoneOrigin\n";
      push (@parsedFile,"; \$ORIGIN      - new origin is $$zoneOrigin\n");
    };
  }
  
  #
  # $INCLUDE
  #
  elsif ($command eq '$INCLUDE') {
    my $file;
    my $origin;
    
    if (@{$directive} > 1) {
      $origin = shift(@{$directive});
      $file   = shift(@{$directive});
    }
    else {
      $origin = $$zoneOrigin; 
      $file   = shift(@{$directive});
    }
    
    # save current so can restore after call to &ParseDBFile.
    my $oldOrigin  = $$zoneOrigin;
    my $oldZoneTTL = $$zoneTTL;


    # TODO - check for infinite loops a includes b which includes a, or a includes a, etc.
    if ($includedfiles{"$inputDir/$file"}) {
      print "Error - Can't include $inputDir/$file - would result in include loop.\n";
    } else {

      $debug && do {
	print "\$INCLUDE # $includedLevel - including $file with origin $origin\n";
	push (@parsedFile,"; \$INCLUDE # $includedLevel - including $file with origin $origin\n");
      };

      $includedLevel++;
      $includedfiles{"$inputDir/$file"}++;
      &ParseDBFile("$inputDir/$file", $origin, $$zoneTTL);
      delete $includedfiles{"$inputDir/$file"};
      $includedLevel--;

      # restore original values.
      $$zoneOrigin = $oldOrigin;
      $$zoneTTL    = $oldZoneTTL;
      
      $debug && do {
	print "Returned from \$INCLUDE # $includedLevel - origin now $$zoneOrigin\n";
	push (@parsedFile,"; Returned from \$INCLUDE # $includedLevel - origin now $$zoneOrigin\n");
      };
    }
  }
  
  #
  # $TTL
  #
  elsif ($command eq '$TTL') { 
    # modifying actual $zoneTTL from ParseDBFile, not a copy.
    $$zoneTTL = shift(@{$directive}); 

    $debug && do {
      print "\$TTL         - new ttl is $$zoneTTL\n";
      push (@parsedFile,"; \$TTL         - new ttl is $$zoneTTL\n");
    };
  }
  
  #
  # $UNKNOWN
  #
  else 
    { 
      print "Error - Ignoring invalid directive $command.\n" 
    }
}


#---------------------------------------------------------------------
#
# GetZoneList - 
#
#---------------------------------------------------------------------

sub GetZoneList() {
  my ($inputDir, $configfile) = @_;
  my (%zone, @zones, $currZone);

  print "Input file is $inputDir/$configfile\n";
  open(CFGFILE, "$inputDir/$configfile") || die "Can't open Conf file $inputDir/$configfile: $!\n";
  while (<CFGFILE>) {
    if (/^\/\//) {next;}
    if (/^zone\s+\"(.+)\"/) {
      $currZone = $1;
#      print $currZone."\n";
      $zone{$currZone}{NAME}=$currZone;
      $zone{$currZone}{DIR} = "forward";
      if ($zone{$currZone}{NAME} =~ /([\d|\.]+)\.IN-ADDR.ARPA/i) {
	$zone{$currZone}{NAME} = lc($zone{$currZone}{NAME});
	$zone{$currZone}{DIR} = "reverse";
      }
    }
    elsif (/\s+type\s+(.+)\s*;/) {
      $zone{$currZone}{TYPE}= uc($1);
    }
    elsif (/\s+file\s+\"(.+)\";/) {
      $zone{$currZone}{FILE}=$1;
    }
    else {}
  }
  close CFGFILE;
  foreach my $key (keys %zone) {
    if ($key) {
      push (@zones, \%{$zone{$key}});
    }
  }
	    
  return \@zones;
}


#---------------------------------------------------------------------
#
# ReverseIP - reverses #.#.#.#
#
#---------------------------------------------------------------------

sub ReverseIP() {
my $IP = $_[0];
my $revIP = join('.', reverse(split(/\./,$IP)));
return $revIP;
}


#---------------------------------------------------------------------
#
# ProcessCmdLine - Get command line options
#
#---------------------------------------------------------------------

sub ProcessCmdLine() {
  use vars qw($opt_o $opt_i $opt_d $opt_h); 


  my $default_outputDir      = "pp";
  my $default_inputDir       = "named"; 
  my $default_debug          = 0;

  my($opt_errs);
  
  $opt_errs = getopts('o:i:dh'); 
  # Doesn't validate parameters: d2xpp -o -i assigns "-i" as outdir
  # but does catch missing parameters: d2xpp -o will return error.
  #[TODO: add some sort of validation for command line input (1st case above)]

  if ($opt_h || !$opt_errs)
    {
      print "\ncommandline: d2xpp.pl [-o outputDir] [-i inDir] [-h]\n";
      print "o - Write output to outputDir (default: pp)\n";
      print "i - Read DNS info from inDir (default: named)\n";
      print "d - Turn on debug output\n";
      print "h - Print Help\n";
      print "\n";
      exit;
    }
  
  my $inputDir    = $opt_i || $default_inputDir;  $inputDir  =~ s/\/$//;
  my $outputDir   = $opt_o || $default_outputDir; $outputDir =~ s/\/$//;
  my $debug       = $opt_d || $default_debug;
  
  return ($inputDir, $outputDir, $debug);
}                                       


