#!/opt/bin/perl5 -w

#=====================================================================
#                                dns2xml
#
# Takes a Ganymede-generated XML file containing DNS info and
# generates DNS zone files from it.
#
# Brian O'Mara
# 11/1999-8/2000
#
# Notes --------------------------------------------------------------
#  Variable naming:
#    Globals_Like_This
#    locals_like_this
#    SubsLikeThis
#  ---
#  If the content of a hash ref (or any variable) is confusing,
#  use the "Data::Dumper" module. Just add a statement to dump
#  out the contents: print Dumper $variable;
#
#=====================================================================

#use diagnostics;
use strict;
use Data::Dumper;
use XML::Parser;
use Getopt::Std;
use File::Copy;

#=====================================================================
# Globals  
#=====================================================================

# Constants

use constant ATTS => 0;

# Config File Options
# (overridden by corresponding command line opts)

use vars qw(
	    $X2D_Config_File  
	    $Input_File          
	    $Output_Dir        
	    $Named_Config_File  
	    $Debug              
	    $Serial_Num_File
	    $Alias_As_Cname
	    $Add_ID_Info
	    $Verbosity
	   );       

# Data Structures

my ($Zones, $Systems, $Serial_Numbers);   

#=====================================================================
# Main  
#=====================================================================
  
&ProcessCmdLine();
&InitializeData();
&LoadSerialNums();
&OutputDnsFiles();
&SaveSerialNums();
exit;


#=====================================================================
# Subroutines 
#=====================================================================

#---------------------------------------------------------------------
#                                                       ProcessCmdLine
#
#     Get command line options and initialize config globals
#
#---------------------------------------------------------------------

sub ProcessCmdLine() {
  $Debug && print "&ProcessCmdLine()\n";

  use vars qw($opt_c $opt_o $opt_i $opt_d $opt_h $opt_a $opt_v $opt_n $opt_s $opt_f); 

  my($opt_errs);
  
  $opt_errs = getopts('c:o:i:d:hav:n:s:bf'); 
  # Doesn't validate parameters: 'dns2xml -o -i' assigns "-i" as outputfile

  if ($opt_h || !$opt_errs)
    {
      print "\nUsage:\n";
      print "$0 [-opt1 [param1]] [-opt2 [param2]] [...]\n";
      print "o - Write output to outputdir\n";
      print "i - Read xml info from infile\n";
      print "a - Aliases are A records\n";
      print "s - Read serial nums from sernumfile";
      print "v - Verbosity. Degree of \"shorthand\" 0-3 used in output\n";
      print "c - xml2dns config file\n";
      print "n - BIND named config file\n";
      print "f - Add ID info\n";
      print "d - Debug output (0=none ... 3=most, default: 0)\n";
      print "h - Print Help\n";
      print "\n";
      exit;
    }

  # Setting globals

  $Output_Dir        = $opt_o || $Output_Dir || "."; $Output_Dir =~ s/\/$//;
  $X2D_Config_File   = $opt_c || "$Output_Dir/xml2dns.conf";

  do "$X2D_Config_File" or die "Can't 'do' config file, $X2D_Config_File: $!";

  $Output_Dir        =  $opt_o || $Output_Dir; $Output_Dir =~ s/\/$//;
  $Input_File        =  $opt_i || "$Output_Dir/$Input_File";
  $Named_Config_File =  $opt_n || "$Output_Dir/$Named_Config_File";
  $Debug             =  $opt_d || $Debug || 0;
  $Serial_Num_File   =  $opt_s || "$Output_Dir/lastserial.dat";
  $Alias_As_Cname    = ($opt_a xor 1);
  $Add_ID_Info       =  $opt_f || 0;
  $Verbosity         =  $opt_v || 3;

  if ($Debug > 1) {
    print "X2D_Config_File    = $X2D_Config_File\n";
    print "Input_File         = $Input_File\n";
    print "Output_Dir         = $Output_Dir\n";
    print "Named_Config_File  = $Named_Config_File\n";
    print "Debug              = $Debug\n";
    print "Serial_Num_File    = $Serial_Num_File\n";
    print "Alias_As_Cname     = $Alias_As_Cname\n";
    print "Add_ID_Info        = $Add_ID_Info\n";
    print "Verbosity          = $Verbosity\n";
  }
}    

#=====================================================================
#
# Output Routines
#
#=====================================================================

#---------------------------------------------------------------------
#                                                       OutputDnsFiles
# 
#---------------------------------------------------------------------

sub OutputDnsFiles {

  # Open a file for all the given zones
  &VisitZoneFiles(\&OpenZone, "FORWARD");
  &VisitZoneFiles(\&OpenZone, "REVERSE");
  
  # extract all the resource records from a system and print them
  foreach my $system (sort {$a->[2][0]{NAME} cmp $b->[2][0]{NAME}} @$Systems){
    &PrintRRs(&GetSystemRRs($system)); 
  } 
  
  # close all the zone files
  &VisitZoneFiles(\&CloseZone, "FORWARD");
  &VisitZoneFiles(\&CloseZone, "REVERSE");  
}


#---------------------------------------------------------------------
#                                                             PrintRRs
# 
#---------------------------------------------------------------------

sub PrintRRs {
  my ($resource_records) = @_;
  
  my %dups = ();
  
  foreach my $rr (@$resource_records) {
    my $domain       = $rr->[0];
    my $domain_dir   = $rr->[1];
    my $resource_rec = $rr->[2];
    
    if (!$dups{$resource_rec}) {
      $dups{$resource_rec}++ ; 
      my $fh = $Zones->{$domain_dir}{$domain}[ATTS]{FILEHANDLE} || "";
      if ($fh) {
	print $fh "$resource_rec\n"; 
      } else {
	print $domain;
	print "No file to print out: $resource_rec\n"; 
      }
    }
  }
}


#---------------------------------------------------------------------
#                                                        GetSystem_RRs
# 
#---------------------------------------------------------------------

sub GetSystemRRs {
  my ($system) = @_;
  
  my $dns_entry  = &GetElement($system,  "DNS");
  my $name       = &GetAtt($dns_entry, "NAME");
  my $domain     = &GetAtt($dns_entry, "DOMAIN");
  
  my $a_rrs          =  &GetA_RRs(      $system, $name, $domain);
  my $mx_rrs         =  &GetMX_RRs(     $system, $name, $domain, $dns_entry);
  my $alias_rrs      =  &GetAlias_RRs(  $system, $name, $domain, $dns_entry);
  my $info_rrs       =  &GetHINFO_RRs(  $system, $name, $domain);
  my $txt_rrs        =  &GetTXT_RRs(    $system, $name, $domain);
  my $rec_rrs        =  &GetRec_RRs(    $system, $name, $domain, $dns_entry);
  my $interface_rrs  =  &GetAllInterfaces_RRs($system, $name, $domain);
  
  my @results = ();
  
  push (@results, @$a_rrs);
  push (@results, @$mx_rrs);
  push (@results, @$alias_rrs);
  push (@results, @$info_rrs);
  push (@results, @$txt_rrs);
  push (@results, @$rec_rrs);
  push (@results, @$interface_rrs);
  
  return \@results;
}


#---------------------------------------------------------------------
#                                                 GetAllInterfaces_RRs
#  
#---------------------------------------------------------------------

sub GetAllInterfaces_RRs {
  my ($system, $name, $domain) = @_;
  my @interfaces = &GetElements($system, "INTERFACE");
  
  my @results = ();
  
  foreach my $interface (@interfaces) {
    my $interface_rrs = &GetInterface_RRs($system, $interface);
    push(@results, @$interface_rrs);
  }
  
  return \@results;
}


#---------------------------------------------------------------------
#                                                     GetInterface_RRs
#
#---------------------------------------------------------------------

sub GetInterface_RRs {
  my ($system, $interface) = @_;
  my @dns_contents = &GetElements($interface, "DNS");
  my @results = ();

  my $dns_entry = &GetElement($system, "DNS");
  my $name      = &GetAtt($dns_entry, "NAME");
  my $domain    = &GetAtt($dns_entry, "DOMAIN");
  
  my $ptr_rrs   = &GetPTR_RRs($interface, $name, $domain, $system);  
  push (@results, @$ptr_rrs);
  
  my $ether_rrs = &GetETHER_RRs($interface, $name, $domain);
  push (@results, @$ether_rrs);

  foreach my $dns_entry (@dns_contents) {
    my $name      = &GetAtt($dns_entry, "NAME");
    my $domain    = &GetAtt($dns_entry, "DOMAIN");
    
    my $a_rrs     = &GetA_RRs(    $interface, $name, $domain);
    my $ether_rrs = &GetETHER_RRs($interface, $name, $domain);
    my $txt_rrs   = &GetTXT_RRs($interface, $name, $domain);   
    my $ptr_rrs   = &GetPTR_RRs($interface, $name, $domain, $system);  
    my $aliases   = &GetAlias_RRs($interface, $name, $domain, $dns_entry);
    my $mxs       = &GetMX_RRs($interface, $name, $domain, $dns_entry);
    my $recs      = &GetRec_RRs($interface, $name, $domain, $dns_entry);   
    
    push (@results, @$a_rrs);
    push (@results, @$ether_rrs);
    push (@results, @$mxs);
    push (@results, @$aliases);
    push (@results, @$txt_rrs);
    push (@results, @$recs);
    push (@results, @$ptr_rrs);
  }  
  return \@results;
}


#---------------------------------------------------------------------
#                                                             GetA_RRs
# 
#---------------------------------------------------------------------

sub GetA_RRs {
  my ($system, $name, $domain) = @_;
  
  my @results = ();
  
  if (defined $system->[ATTS]{"PTRTYPE"}) {
    my @system_ips = ();
    foreach my $interface (&GetElements($system, "INTERFACE")) {
      my @interface_ips = &GetElements($interface, "IP");
      foreach my $ip (@interface_ips) {
	push(@system_ips, &GetAtt($ip, "ADDR"));
      }
    }
    
    foreach my $ip (@system_ips) {
      &AddRR("A", $ip, $name, $domain, \@results);
    }
    
  }
  else {
    foreach my $ip (&GetElements($system, "IP")) {
      &AddRR("A", &GetAtt($ip, "ADDR"), $name, $domain, \@results);
    }
  }
  @results = sort {$a->[1] cmp $b->[1]}@results;
  return \@results;
}


#---------------------------------------------------------------------
#                                                            GetMX_RRs
# 
#---------------------------------------------------------------------

sub GetMX_RRs {
  my ($system, $name, $domain, $dns_entry) = @_;
  
  my @mx_hosts = &GetElements($dns_entry, "MX_HOST");
  
  if (!@mx_hosts) {
    my $default_mx = &GetElement($Zones->{"FORWARD"}{$domain}, "DEFAULT_MX");
    if (!@$default_mx) {
      $default_mx = &GetElement($Zones->{"FORWARD"}{"FORWARD"}, "DEFAULT_MX"); 
    }
    @mx_hosts = &GetElements($default_mx, "MX_HOST");
  }
  
  my @results = ();
  
  foreach my $mx_host (@mx_hosts) {
    my $cost = &GetAtt($mx_host, "COST");
    my $host = &GetElement($mx_host, "0");
    
    if ($cost ne "") { 
      &AddRR("MX", "$cost $host", $name, $domain, \@results);
    }
 
  }
  @results = sort {$a->[1] cmp $b->[1]}@results;
  return \@results;  
}


#---------------------------------------------------------------------
#                                                         GetAlias_RRs
# 
#---------------------------------------------------------------------

sub GetAlias_RRs {
  my ($system, $name, $domain, $dns_entry) = @_;

  my @results = ();

  foreach my $alias (&GetElements($dns_entry, "ALIAS")) {
    my $a_name   = &GetAtt($alias, "NAME");
    my $a_domain = &GetAtt($alias, "DOMAIN");

    if ($Alias_As_Cname) { 
      &AddRR("CNAME", &ToFQDN($name, $domain), $a_name, $a_domain, \@results);
    }
    else {
      my $mx_rrs = &GetMX_RRs($system, $a_name, $a_domain, $dns_entry);
      my $a_rrs  = &GetA_RRs( $system, $a_name, $a_domain);
      
      print Dumper $mx_rrs;
      print Dumper $a_rrs;
      push (@results, @$a_rrs);
      push (@results, @$mx_rrs);
    }
  }
  
  @results = sort {$a->[1] cmp $b->[1]}@results;
  return \@results;
}


#---------------------------------------------------------------------
#                                                         GetHINFO_RRs
# 
#---------------------------------------------------------------------

sub GetHINFO_RRs {
  my ($system, $name, $domain) = @_;
  my @results = ();
  
  if ($Add_ID_Info) { 
    
    my $info_contents = &GetElement($system, "INFO");
    my $cpu = &GetAtt($info_contents, "CPU");
    my $os  = &GetAtt($info_contents, "OS");
    
    if (($cpu ne "N/A") || ($os ne "N/A")) { 
      &AddRR("HINFO", "\"$cpu\" \"$os\"", $name, $domain, \@results);
    }
    @results = sort {$a->[1] cmp $b->[1]}@results;
  }
  
  return \@results;
}

#---------------------------------------------------------------------
#                                                           GetTXT_RRs
# 
#---------------------------------------------------------------------

sub GetTXT_RRs {
  my ($system, $name, $domain) = @_;
  my @results = ();
  # todo....
  return \@results;
}

#---------------------------------------------------------------------
#                                                           GetRec_RRs
# 
#---------------------------------------------------------------------

sub GetRec_RRs {
  my ($system, $name, $domain, $dns_entry) = @_;

  my $records_contents = &GetElement($dns_entry, "RECORDS");
  my @records = &GetElements($records_contents, "RECORD");

  my @results = ();
  foreach my $record (@records) {
    my $text = &GetElement($record, "0");    
    &AddRR("REC", "$text",  $name, $domain, \@results);
  }
  @results = sort {$a->[1] cmp $b->[1]}@results;
  return \@results;
}	     

#---------------------------------------------------------------------
#                                                           GetPTR_RRs
# 
#---------------------------------------------------------------------

sub GetPTR_RRs {
  my ($interface, $name, $domain, $system) = @_;
  my @results = ();
  
  my $system_ptrtype = &GetAtt($system, "PTRTYPE");
  
  if ($system_ptrtype == 1) { 
    foreach my $ip (&GetElements($interface, "IP")) {
      my $ptr_name = &ReverseIP((&GetAtt($ip,"ADDR"))) . ".in-addr.arpa.";
      my ($name, $domain) = &FromFQDN($ptr_name, "REVERSE");
      my $dns_entry = &GetElement($system, "DNS"); 
      my $sys_name = &GetAtt($dns_entry, "NAME");
      my $sys_domain = &GetAtt($dns_entry, "DOMAIN");
      
      &AddRR("PTR", "$sys_name.$sys_domain", $name, $domain, \@results);
    }
  }
  elsif ($system_ptrtype == 2) {
    foreach my $ip (&GetElements($interface, "IP")) {
      my $ptr_name = &ReverseIP((&GetAtt($ip,"ADDR"))) . ".in-addr.arpa.";
      my ($name, $domain) = &FromFQDN($ptr_name, "REVERSE");
      foreach my $dns_entry (&GetElements($interface, "DNS")) {
	my $dns_name = &GetAtt($dns_entry, "NAME");
	my $dns_domain = &GetAtt($dns_entry, "DOMAIN");
	  
	&AddRR("PTR", "$dns_name.$dns_domain", $name, $domain, \@results);
      }
    }
  }
  else { # ptrtype 0 or not 1 or 2
    
    foreach my $ip (&GetElements($interface, "IP")) {
      foreach my $ptr (&GetElements($ip, "PTR")) {
	my $ptr_name = &ReverseIP((&GetAtt($ip,"ADDR"))) . ".in-addr.arpa.";
	my ($name, $domain) = &FromFQDN($ptr_name, "REVERSE");
	&AddRR("PTR", &GetAtt($ptr, "NAME"), $name, $domain, \@results);
      }
    }
  }
    
  @results = sort {$a->[1] cmp $b->[1]}@results;
  return \@results;
}

#---------------------------------------------------------------------
#                                                         GetEther_RRs
# 
#---------------------------------------------------------------------

sub GetETHER_RRs { 
  my ($interface, $name, $domain) = @_;
  my @results = ();
  
  if ($Add_ID_Info) {
    
    foreach my $ether (&GetElements($interface, "ETHER")) {
      &AddRR("TXT","\"".&GetAtt($ether, "ADDR")."\"",$name,$domain,\@results);
    }
    
    @results = sort {$a->[1] cmp $b->[1]}@results;
  }
  return \@results;
}


#---------------------------------------------------------------------
#                                                                AddRR
# 
#---------------------------------------------------------------------

sub AddRR() {
  my ($type, $data, $name, $domain, $RRs) = @_;
  
  my $zone_direction = ($Zones->{REVERSE}{$domain} ? "REVERSE" : "FORWARD");

  if ($type eq "REC") { 
    push(@$RRs, [$domain, $zone_direction, "$data"]); 
  }
  else {
    my $owner = &AdjustVerbosity($name, $domain);
    if ($type eq "CNAME") { $data = &AdjustVerbosity($data, $domain); }   
    push(@$RRs, [$domain, $zone_direction, "$owner\tIN\t$type\t$data"]);
  }
}


#---------------------------------------------------------------------
#                                                      AdjustVerbosity
# 
#---------------------------------------------------------------------

sub AdjustVerbosity {
  my ($name, $domain) = @_;
  my $owner = &ToFQDN($name, $domain);
  if ($Verbosity == 0) {
    if ($owner eq $domain) { 
      $owner = "@"; 
    }
    elsif ($owner =~ /^(\S+)\.$domain/) { 
      $owner = $1; 
    }
  }
  
  elsif ($Verbosity == 1) {
    if ($owner eq "@") { $owner = $domain; }
    elsif ($owner =~ /^(\S+)\.$domain/) { $owner = $1; } 
    else {$owner = $domain;} 
  }
  
  else { 
    if ($owner eq "@") { $owner = $domain; }
    $owner = &ToFQDN($name, $domain);
  }
}


#=====================================================================
#
# Data Access Routines
#
#=====================================================================

#---------------------------------------------------------------------
#                                                           GetElement
# 
#---------------------------------------------------------------------

sub GetElement {
  my ($parent, $child) = @_;
  my $result = [];
  if (!defined($parent)) {$parent = [];}
  
  for (my $i = 0; $i < @$parent; $i++) {
    if ($parent->[$i] eq $child) {
      $result = $parent->[$i+1];
      last;
    }
  }
  return $result;
}


#---------------------------------------------------------------------
#                                                          GetElements
# 
#---------------------------------------------------------------------

sub GetElements {
  my ($parent, $child) = @_;
  my @results = ();
  if (!defined($parent)) {$parent = [];}
  
  for (my $i = 0; $i < @$parent; $i++) {
    if ($parent->[$i] eq $child) {
      push (@results, $parent->[$i+1]);
    }
  }
  return @results;
}

#---------------------------------------------------------------------
#                                                               GetAtt
# 
#---------------------------------------------------------------------

sub GetAtt {
  my ($element, $att) = @_;
  $att = $element->[ATTS]{$att};
  return $att;
}

#---------------------------------------------------------------------
#                                                           GetContent
# 
#---------------------------------------------------------------------

sub GetContent {
  my ($element) = @_;
  my $content = $element->{"0"} || "";
  return $content;
}

#---------------------------------------------------------------------
#                                                               DoChar
#
# Overrides XML::Parser's character handling routine to strip 
# whitespace.
#
#---------------------------------------------------------------------

sub DoChar {
  my $expat = shift;
  my $text = shift;
  my $clist = $expat->{Curlist};
  my $pos = $#$clist;
  
  $text =~ s/^\s*$//;
  if ($text) {
    if ($pos > 0 and $clist->[$pos - 1] eq '0') {
      $clist->[$pos] .= $text;
    } else {
      push @$clist, 0 => $text;
    }
  }   
}


#=====================================================================
#
# Utility Routines
#
#=====================================================================

#---------------------------------------------------------------------
#                                                               ToFQDN
# 
#     Takes a domain name and if it doesn't end in '.', adds the
#     necessary zone ending to it.
#
#---------------------------------------------------------------------

sub ToFQDN {
  my (
      $data, # domain name to transform  
      $zone, # zone to append
     ) = @_;

  ($Debug > 2) && print "&ToFQDN()\n";

  my $fqdn_data = $data;


  # ends in "." : must already be fqdn
  if ($data =~ /\.$/) {
    $fqdn_data = $data;
  }

  # IP - #.#.#.# : reverse & add in-addr
  elsif ($data =~ /^\d+\.\d+\.\d+\.\d+$/) {
    $fqdn_data = &ReverseIP($data).".in-addr.arpa.";
  }

  # add domain to dns_name
  else { 
    $fqdn_data = $data . "." . $zone; 
  }
  
  return $fqdn_data;
}        


#---------------------------------------------------------------------
#                                                             FromFQDN
# 
#     Takes a fqdn and returns it as (name, domain)
#
#---------------------------------------------------------------------

sub FromFQDN {
  my (
      $dns_name,  # fqdn to transform into name, domain  
      $direction, # FORWARD or REVERSE zone
     ) = @_;

  ($Debug > 2) && print "&FromFQDN()\n";

  my $name        = $dns_name;
  my $domain      = $dns_name;

  if (!defined $Zones->{$direction}{$dns_name}) {
    my @name_bits    = ();
    my @domain_bits  = split(/\./,$dns_name);
    my $domain_found = 0;

    while (!$domain_found && @domain_bits) {
      push(@name_bits, (shift @domain_bits));
      
      $domain = join('.',@domain_bits) . ".";
      $name   = join('.',@name_bits);
#     print "$name    $domain\n";
      
      if (defined $Zones->{$direction}{$domain}) { $domain_found = 1; }
    }
  }
  return ($name, $domain);
}



#---------------------------------------------------------------------
#                                                            ReverseIP
#
#     Takes an IP and reverses the order of the numbers.
#     (1.2.3.4 returns 4.3.2.1)     
#
#---------------------------------------------------------------------

sub ReverseIP {
  my ($IP) = @_; # IP to reverse
  ($Debug > 2) && print "&ReverseIP()\n";

  $IP =~ s/\s+//g;
  my $revIP = join('.', reverse(split(/\./,$IP)));
  return $revIP;
}  


#=====================================================================
#
# Data & Initialization
#
#=====================================================================


sub InitializeData {
  my $parser = new XML::Parser(Style =>'Tree', Handlers =>{Char => \&DoChar});
  my $tree    = $parser->parsefile($Input_File);
#  print Dumper $tree; exit;
  my $gany_dns_contents = &GetElement($tree, "GANY_DNS");

  my $systems_contents  = &GetElement($gany_dns_contents, "SYSTEMS");
  my @systems = &GetElements($systems_contents, "SYSTEM");

  my $forward_zone_contents = &GetElement($gany_dns_contents, "FORWARD_ZONES");
  my @fzones  = &GetElements($forward_zone_contents, "ZONE");

  my $reverse_zone_contents = &GetElement($gany_dns_contents, "REVERSE_ZONES");
  my @rzones  = &GetElements($reverse_zone_contents, "ZONE");

  foreach (@fzones) { $Zones->{FORWARD}{&GetAtt($_, "NAME")} = $_; }
  foreach (@rzones) { $Zones->{REVERSE}{&GetAtt($_, "NAME")} = $_; }

  $Systems = \@systems;
}

#=====================================================================
#
# Serial Number Routines
#
#=====================================================================

#---------------------------------------------------------------------
#                                                       LoadSerialNums
# 
# 
#---------------------------------------------------------------------
sub LoadSerialNums {
  
  if (-e $Serial_Num_File) {
    open (SERIALS, "$Serial_Num_File") or 
      die "Can't open serial number file: $Serial_Num_File\n";
    
    while (<SERIALS>) {
      chomp;
      next if (/^;.*/);
      if (/^(\S+)\s+(\d+)$/) {
	my $zone_file_name = $1;  
	my $serial_number  = $2;
	$Serial_Numbers->{$zone_file_name} = $serial_number;
      }
    }
    close SERIALS;
  }
  else {
    &VisitZoneFiles(\&GenSerialNums, "FORWARD");
    &VisitZoneFiles(\&GenSerialNums, "REVERSE");
  }
}  

sub GenSerialNums {
  my ($zone, $zone_direction) = @_;
  my $file = $Zones->{$zone_direction}{$zone}[ATTS]{FILE} || "";
  $file && &GetSerialNum($file);
}

#---------------------------------------------------------------------
#                                                       SaveSerialNums
# 
# Write out the appropriate DNS serial numbers for the zone files.
# 
#---------------------------------------------------------------------
sub SaveSerialNums {
  
  # copy currfile to currfile.old if exists...
  if (-e "$Serial_Num_File") {
    copy("$Serial_Num_File", "$Serial_Num_File.old") 
      or die "Can't copy $Serial_Num_File to $Serial_Num_File.old\n";
  }
  
  # write out new currfile  
  open (SERIALS, ">$Serial_Num_File") 
    or die "Can't open $Serial_Num_File\n";
  
  print SERIALS ";\n; Zone Serial Numbers as of ". localtime(time) ."\n;\n";
  
  foreach my $file (sort keys %{$Serial_Numbers}) {
    print SERIALS "$file\t\t" . $Serial_Numbers->{$file} . "\n";
  }
  
  close SERIALS;
}


#---------------------------------------------------------------------
#                                                         GetSerialNum
#
# Generates the appropriate DNS serial number for the zone file header. 
# 
#---------------------------------------------------------------------
sub GetSerialNum {
  my ($zone_file) = @_; 
  my @days = (0,31,28,31,30,31,30,31,31,30,31,30,31);

  # Base increment is 00. (Gets bumped to 01 before written out.) 
  my $inc = "00";
  my $date = "";
  my $curr_serial_num = "";
  
  # if has current serial number, grab current info.
  if (defined ($Serial_Numbers->{$zone_file})) { 
    $curr_serial_num = $Serial_Numbers->{$zone_file};       
    $date = substr($curr_serial_num, 0,8);      
    $inc  = substr($curr_serial_num,-2,2);     
  }

  # generate today's date...
  my ($month_day, $month, $year) = (localtime(time))[3..5];
  
  # 4-digit year
  $year = $year + 1900;
  
  # 2-digit month
  $month = $month + 1;
  if (length($month)     < 2) {$month     = "0$month";}
  if (length($month_day) < 2) {$month_day = "0$month_day";}
  
  my $todays_date = "$year$month$month_day";
  
  $curr_serial_num=$todays_date . (($date eq $todays_date) ? "$inc" : "00"); 
  
  $Serial_Numbers->{$zone_file} = $curr_serial_num;
  
  # Increment it for write-out
  ($year, $month, $month_day, $inc) = 
    $Serial_Numbers->{$zone_file} =~ /(\d\d\d\d)(\d\d)(\d\d)(\d\d)/;
  
  # do the increment and associated roll-overs   
  $inc++;
  if ($inc > 99) {
    $inc = 0; $month_day++;
    if ($month_day > $days[$month]) {
      $month_day = 1; $month++;
      if ($month > 12) {
	$month = 1; $year++;
	if ($year > 9999) {
	  $year = 0; $month = 1; $month_day = 1; $inc = 0;
	}
      }
    }
  } 
  # make sure YYYYMMDDII
  my $new_serial_num = sprintf("%4d%2d%2d%2d",$year,$month,$month_day,$inc); 
  $new_serial_num =~ tr/ /0/; # turn 'space' padding into '0' padding
  $Serial_Numbers->{$zone_file} = $new_serial_num; 
}


#---------------------------------------------------------------------
#                                                       VisitZoneFiles
#
#---------------------------------------------------------------------

sub VisitZoneFiles {
  my ($sub_ref, $zone_direction) = @_;
  foreach my $zone (keys %{$Zones->{$zone_direction}}) {
    ($Zones->{$zone_direction}{$zone}[ATTS]{TYPE} ne "DEFAULT") && 
      &$sub_ref($zone, $zone_direction); 
  }
}


#---------------------------------------------------------------------
#                                                             OpenZone
# 
#---------------------------------------------------------------------

sub OpenZone {
  my ($domain, $dir) = @_;
  no strict 'refs';
  my @results = ();

  my $fh;
  # Determine if file already opened. If not open it and print header.
  if (defined($Zones->{$dir}{$domain}[ATTS]{FILEHANDLE})) {
    $fh = $Zones->{$dir}{$domain}[ATTS]{FILEHANDLE};
  }
  else {
    $fh = $Zones->{$dir}{$domain}[ATTS]{FILE};
    open($fh, ">$Output_Dir/$fh") 
      or die "Can't open zonefile $Output_Dir/$fh, $! \n";
    
    # store reference to file so we can grab it next time.    
    $Zones->{$dir}{$domain}[ATTS]{FILEHANDLE} = \*$fh;
    
    print $fh ";\n; Origin added to names not ending in a dot: $domain\n;\n";
    # get zone SOA info & print to file
    my $zone_atts    = $Zones->{$dir}{$domain}[ATTS];
    my $default_atts = $Zones->{$dir}{$dir}[ATTS];

    my $refresh  = $zone_atts->{REFRESH}  || $default_atts->{REFRESH};
    my $retry    = $zone_atts->{RETRY}    || $default_atts->{RETRY};
    my $expire   = $zone_atts->{EXPIRE}   || $default_atts->{EXPIRE};
    my $minttl   = $zone_atts->{MINTTL}   || $default_atts->{MINTTL};
    my $origin   = $zone_atts->{HOST}     || $default_atts->{HOST};
    my $contact  = $zone_atts->{MAILADDR} || $default_atts->{MAILADDR};
    
    my $serialnum = &GetSerialNum($fh);
    
    my $rdata =  "$origin\t$contact (\n";
    $rdata .= "\t" x 4 . "$serialnum \t ; Serial (YYYYMMDDNN)\n";  
    $rdata .= "\t" x 4 . "$refresh \t\t ; Refresh\n";
    $rdata .= "\t" x 4 . "$retry \t\t ; Retry\n";
    $rdata .= "\t" x 4 . "$expire \t\t ; Expire\n";
    $rdata .= "\t" x 4 . "$minttl )\t\t ; Minimum TTL";

    &AddRR("SOA", $rdata, $domain, $domain, \@results);
    
    # get zone nameserver info & print to file    
    my $ns_contents = &GetElement($Zones->{$dir}{$domain}, "NAMESERVERS");
    if (!@$ns_contents) { 
      $ns_contents = &GetElement($Zones->{$dir}{$dir}, "NAMESERVERS");
    }
    
    if (defined(@$ns_contents)) {
      my $text = "\n;\n; Name Servers\n;\n";
      &AddRR("REC", $text, $domain, $domain, \@results);

      my @ns_hosts = &GetElements($ns_contents, "HOST");

      foreach my $server (@ns_hosts) {
	&AddRR("NS", &GetElement($server, "0") , $domain, $domain, \@results);
      }
      
      
      # get zone nameserver info & print to file    
      my $rec_contents = &GetElement($Zones->{$dir}{$domain}, "RECORDS");
      
      if (defined(@$rec_contents)) {
	my $text = "\n;\n; Data from xml <RECORDS> for zone $domain\n;\n";
	&AddRR("REC", $text, $domain, $domain, \@results);
	
	my @records = &GetElements($rec_contents, "RECORD");
	
	foreach my $record (@records) {
	  &AddRR("REC",&GetElement($record, "0"),$domain,$domain,\@results);
	}
      }
    }
    my $text = "\n;\n; Information determined from Ganymede database\n;\n";
    &AddRR("REC", $text, $domain, $domain, \@results);
  } 
  &PrintRRs(\@results);
}

#---------------------------------------------------------------------
#                                                            CloseZone
# 
#---------------------------------------------------------------------

sub CloseZone {
  my ($domain, $dir) = @_;
  my $fh = $Zones->{$dir}{$domain}[ATTS]{FILEHANDLE} || "";
  if ($fh) {close $fh or die "Can't close filehandle $fh, $!";}
}

